open Term
open Monad.MonadList

(*************************************************)
(*                UNIFICATION                    *)
(*************************************************)

(** basic matching **)
let rec var_clash (x,t) = function
    | []          -> false
    | (y,s) :: ys when x = y
                  -> t <> s || var_clash (x,t) ys
    | _ :: ys     -> var_clash (x,t) ys

let consist sigma =
  List.for_all
    (fun s -> not (var_clash s sigma))
    sigma

(* Notice: Exception List.combine *)
let rec match_with sub p s = match p,s with
  | V x, _ -> (x,s) :: sub
  | F (f,ps), F (g,ss) when f = g ->
      let* (a,b) = List.combine ps ss in
      match_with sub a b
  | _      -> failwith "variable clash"

let matcher p s =
  let matched = match_with [] p s in
  if consist matched then
    matched
  else
     failwith "variable clash"

let rec instance_of s p =
  try
    ignore (matcher p s); true
  with _ ->
    false

let matchers trs s =
  let* (l,r) = trs in
  guard (instance_of s l) >>= fun () ->
  [(l,r), (matcher l s)]

let rec variant s t = instance_of s t && instance_of t s

let variant_rule (l1,r1) (l2,r2) =
  let m x y = F ("",[x;y]) in
  variant (m l1 r1) (m l2 r2)


(****** unification *****)

let update_elist s ts =
  List.map
    (fun (a,b) -> (substitute [s] a, substitute [s] b))
    ts
let update_sigma s sigma =
  List.map
    (fun (x,t) -> (x, substitute [s] t))
    sigma

let not_mem a list = not (List.mem a list)

let rec unification sigma = function
  | []      -> sigma
  | e :: es ->
    let next s =
      unification
        (s :: update_sigma s sigma)
        (update_elist s es) in
  match e with
    | t, u   when t = u
      -> unification sigma es
    | V x, t when not_mem x (variables t)
      -> next (x,t)
    | t, V x when not_mem x (variables t)
      -> next (x,t)
    | F (f,ts1), F (g,ts2) when f = g
      -> unification
          sigma
          (List.combine ts1 ts2  @  es)
    | _ -> failwith "not unifiable"

let mgu es = unification [] es

let unifiable es =
  try
    ignore (mgu es); true
  with _ -> false


(** TCAP function **)
let rec tcap trs = function
  | V x      -> new_var ()
  | F (f,ts) -> let u = F (f, List.map (tcap trs) ts) in
    if List.exists
      (fun (l,_) -> unifiable [l,u])
      trs
    then
      new_var ()
    else
      u


(*************************************************)
(**************** Critical Pair ******************)
(*************************************************)
let rec simple_cp = function
  | []         -> []
  | (a,b)::cps when variant a b
               -> simple_cp cps
  | (a,b)::cps -> let cntx ts = F ("_",ts) in
    (a,b) :: List.filter
              (fun (c,d) -> not (variant (cntx [a;c]) (cntx [b;d])))
              (simple_cp cps)

(** functions about overlap **)

let is_overlap ((a,b),p,(c,d)) = 
  let var (x,y) = variables x @ variables y in
  let at = subterm_at p c in
  try 
    Listset.inter (var (a,b)) (var (c,d)) = [] (* without common variables *)
      &&
    not (is_variable at)
      &&
    unifiable [a, at] (* if position p is invalid, then exception arises *)
    (*
      &&
    not (p = [] && variant (F ("",[a;b])) (F ("",[c;d]))) (* if p is root, then rules are not variant *)
    *)
  with Failure _ -> false

(* return renamed overlaps if TRSs contain common vars *)
let overlaps r s =
  let s' =
    if Listset.intersect (Trs.variables r) (Trs.variables s) then
      Trs.rename s
    else
      s in
  let* a = r in
  let* b = s' in
  let (l2,_) = b in
  let* p = function_positions l2 in
  guard (is_overlap (a, p, b)) >>= fun () ->
  [a, p, b]


(** Critical Pair **)
let cp_overlap laps =
  let* ((l1,r1),p,(l2,r2)) = laps in
  guard (unifiable [l1, subterm_at p l2]) >>= fun () ->
  guard (p <> [] || not (Rule.variant (l1,r1) (l2,r2))) >>= fun () ->
  let s = mgu [l1, subterm_at p l2] in
  [replace (substitute s l2) (substitute s r1) p, (* rewrite inner redex *)
  p, (* inner redex position *)
  substitute s l2, (* root term *)
  substitute s r2 (* rewrite outer redex *)
  ]

(* remove variant critical pairs *)
let c_pp laps =
  Listx.unique_by
    (fun (l1,p1,_,r1) (l2,p2,_,r2) ->
      variant_rule (l1,r1) (l2,r2))
    (cp_overlap laps)

(** Critical Peak **)

(* critical peaks for different TRSs *)
let cpp2 r s = c_pp (overlaps r s)

(* lhs is rewritten by root position *)
let cpp2r r s =
  List.map (fun (t,p,s,u) -> (u,p,s,t)) (cpp2 s r)


(* <-R- ><| -S-> U <-R- |>< -S-> *)
let cpp2_all r s = Listset.union (cpp2 r s) (cpp2r r s)

let cpp2_in r s =
  overlaps r s
  |> List.filter (fun (_,p,_) -> p <> [])
  |> c_pp

let cpp2_out r s =
  overlaps r s
  |> List.filter (fun (_,p,_) -> p = [])
  |> c_pp

(* critical peaks for the TRS *)
let cpp rs = cpp2 rs rs
let cpp_in rs = cpp2_in rs rs
let cpp_out rs = cpp2_out rs rs


(** Critical Pair **)
(* sorted critical pairs *)
let c_p xs = List.map (fun (l,_,_,r) -> (l,r)) xs

(* critical pairs for different trs's *)
let cp2 r s = c_p (cpp2 r s)

(* lhs is rewritten by root position *)
let cp2r r s = c_p (cpp2r r s)

(* <-R- ><| -S-> U <-R- |>< -S-> *)
let cp2_all r s = c_p (cpp2_all r s)

(* critical pairs for a trs *)
let cp trs = cp2 trs trs

let cp2_in r s = c_p (cpp2_in r s)
let cp2_out r s = c_p (cpp2_out r s)

let cp_in rs = cp2_in rs rs
let cp_out rs = cp2_out rs rs

(* remove identical critical pair *)
let rm_ident1     = List.filter (fun (x,_,_,y) -> x <> y)
let cpp2_ r s     = cpp2 r s     |> rm_ident1
let cpp2r_ r s    = cpp2_ r s    |> rm_ident1
let cpp2_all_ r s = cpp2_all r s |> rm_ident1
let cpp2_in_ r s  = cpp2_in r s  |> rm_ident1
let cpp2_out_ r s = cpp2_out r s |> rm_ident1
let cpp_ r        = cpp2 r r     |> rm_ident1

let rm_ident2     = List.filter (fun (x,y) -> x <> y)
let cp2_ r s     = cp2 r s     |> rm_ident2
let cp2r_ r s    = cp2_ r s    |> rm_ident2
let cp2_all_ r s = cp2_all r s |> rm_ident2
let cp2_in_ r s  = cp2_in r s  |> rm_ident2
let cp2_out_ r s = cp2_out r s |> rm_ident2
let cp_ r        = cp2 r r     |> rm_ident2

let print_criticalpair ppf (t,u) =
  Format.fprintf ppf "%a <--x--> %a" Term.print t Term.print u

let print_criticalpeak ppf (t,s,u) =
  Format.fprintf ppf "%a <-- %a --> %a" Term.print t Term.print s Term.print u

let print_criticalpeak_at ppf (t,p,s,u) =
  Format.fprintf ppf "%a <-%a- %a --> %a"
    Term.print t
    Term.print_position p
    Term.print s
    Term.print u

let print_criticalpairs ppf cps =
  Format.fprintf ppf "%a"
    (Formatx.print_list
      print_criticalpair
      "\n")
  cps

let print_criticalpeaks ppf cps =
  Format.fprintf ppf "%a"
    (Formatx.print_list
      print_criticalpeak
      "\n")
  cps

let print_criticalpeaks_at ppf cps =
  Format.fprintf ppf "%a"
    (Formatx.print_list
      print_criticalpeak_at
      "\n")
  cps

(** Simultaneous Critical Pairs **)

(* sort redex patterns by ascending order of position (root is smallest) *)
let sort_patterns_asc ps =
  let compare (p,_) (q,_) = Stdlib.compare (List.length p) (List.length q) in
  List.sort compare ps

(* sort redex patterns by descending order of position (root is greatest) *)
let sort_patterns_des ps =
  List.rev (sort_patterns_asc ps)

let rec unconnected_with (l,r) ss =
  List.for_all
    (fun (p,_) -> not (List.mem p (function_positions l)))
    ss

let rec unconnected ss =
  unconnected_aux (sort_patterns_asc ss)
and unconnected_aux = function
    [] -> true
  | (_,lr)::ss -> unconnected_with lr ss && unconnected_aux ss

let add_position p patterns =
  List.map (List.map (fun (q,a) -> (p@q,a))) patterns

let power_product xss =
  Listx.power xss
  |> List.filter ((<>)[]) 
  (* combine unconnected redex patterns *)
  |> List.map (fun u -> List.map List.concat (Listx.pi u))
  (* flatten set *)
  |> List.concat

(* generate list of all overlapping simultaneous set of t
   i.e. list of list of redex patterns.
 *)
let rec simultaneous_sets rs = function
    V x -> []
  | F (f,ts) as t ->
     if Listset.inter (Term.variables t) (Trs.variables rs) <> [] then
       simultaneous_sets_aux rs (Term.rename t)
     else
       simultaneous_sets_aux rs t
and simultaneous_sets_aux rs = function
    V x -> []
  | F (f,ts) as t ->
     let subterm_sset (i,t) = add_position [i] (simultaneous_sets rs t) in
     (* list of all simultaneous patterns on subterms *)
     let sss = List.map subterm_sset (Listx.index ts) in
     let root_patterns = List.filter (fun (l,_) -> unifiable [(t,l)]) rs
                         |> List.map (fun rl -> [[], rl]) in
     if root_patterns = [] then
       power_product sss
     else
       power_product (root_patterns :: sss)
       |> List.filter unconnected

(* assume that
   (1) t is linear, (2) patterns is simultaneous,
   and (3) t overlap l at p for all (p,(l,_)) in patterns *)
let rec sumup_patterns t patterns =
  List.fold_left
    (fun u (p,(l,_)) -> substitute (mgu [l,subterm_at p u]) u)
    t
    patterns

let remove_variant_overlaps patterns =
  List.filter
    (fun (x,p,_,lr1) ->
      match x with [q,lr2] -> not (p = q && Rule.variant lr1 lr2) | _ -> true)
    patterns


(** Simultaneous critical pairs *)
(* [Rule] -> Rule -> [[Pos,Rule],Pos,Rule] *)
let simultaneous_overlap_outer rs (l,r) =
  simultaneous_sets rs l
  |> Listx.concat_map
       (fun patterns ->
           [(patterns, [], sumup_patterns l patterns, (l,r))])
  |> remove_variant_overlaps

let simultaneous_overlaps_outer r s =
  let rr,ss =
    if Listset.inter (Trs.variables r) (Trs.variables s) = []
    then r,s
    else Trs.rename r, Trs.rename s in
  Listx.concat_map (simultaneous_overlap_outer rr) ss

(* [Rule] -> Rule -> [[Pos,Rule],Pos,Term,Rule] *)
let simultaneous_overlap_inner rs (l,r) =
  let mem (p,lr) patterns =
    List.exists (fun (q,lr2) -> p = q && Rule.variant lr lr2) patterns
  in
  overlaps [l,r] rs
  |> List.filter (fun (_,p,_) -> p <> [])
  |> Listx.concat_map (fun (_,p,(l2,r2)) ->
         (* lookup with simultaneous set of overlay redex t *)
         let t = substitute (mgu [l,subterm_at p l2]) l2 in
         List.map
           (fun patterns -> patterns, p, sumup_patterns t patterns ,(l,r))
           (List.filter (mem ([],(l2,r2))) (simultaneous_sets rs t)))
  |> remove_variant_overlaps

let simultaneous_overlaps_inner r s =
  let rr,ss =
    if Listset.inter (Trs.variables r) (Trs.variables s) = []
    then r,s
    else Trs.rename r, Trs.rename s in
  Listx.concat_map (simultaneous_overlap_inner rr) ss

let rec rewrite_redex_patterns t patterns =
  (* rewrite term t at botom to top to ensure /complete/ development *)
  List.fold_left
    (fun u (p,lr) -> Rewriting.rewrite_with lr p u)
    t
    (sort_patterns_des patterns)

(* [Rule] -> [Rule] -> [[Pos,Rule],Pos,Term,Rule] *)
let simultaneous_overlaps r s =
  simultaneous_overlaps_inner r s @ simultaneous_overlaps_outer r s
  
let simultaneous_cp2 r s =
  Listx.unique
    (List.map
       (fun (patterns,p,t,lr) ->
         rewrite_redex_patterns t patterns, Rewriting.rewrite_with lr p t)
       (simultaneous_overlaps r s))

let simultaneous_cp r = simultaneous_cp2 r r

let print_patterns ppf patterns =
  Format.fprintf ppf "[%a]"
    (Formatx.print_list
      (fun ppf (p,rl) -> Format.fprintf ppf "%a with %a@." print_position p Rule.print rl)
      ";")
  patterns

let print_simultaneous_overlaps ppf laps =
  Format.fprintf ppf "%a"
    (Formatx.print_list
      (fun ppf (ps,p,t,rl) ->
        Format.fprintf ppf "%a <-o-- %a -->\n%a with %a@."
          print_patterns ps
          Term.print t
          print_position p
          Rule.print rl)
      "\n")
  laps

(** Parallel critical pairs **)

let rec parallel p q = match (p,q) with
    i::p', j::q' -> i <> j || parallel p' q'
  | _            -> false
                
let parallel_positions positions =
  List.for_all
    (fun p -> List.for_all
                (fun q -> p = q || parallel p q)
                positions)
    positions

let parallel_overlaps_aux r s =
  let pos xs = List.map fst xs in
  simultaneous_overlaps r s
  |> List.filter (fun (patterns,p,t,rl) ->
         match patterns with
           [[],_] -> true
         | _      -> p = [] && parallel_positions (pos patterns))
  |> List.map (fun (patterns,p,t,rl) ->
         if p = [] then (patterns,t,rl) 
         else
           match patterns with
             [[],rl2] -> ([p,rl],t,rl2)
           | _        -> invalid_arg "argument is not parallel overlap")

let parallel_overlaps r s =
  let make_peak t lr patterns =
    (rewrite_redex_patterns t patterns,
    List.map (fun (x,y) -> (y,x)) patterns,
    t,
    (lr,[]),
    Rewriting.rewrite_with lr [] t) in
  Listx.unique
    (Listx.concat_map
       (fun (patterns,t,lr) ->
         (* t and redex patterns may have common variables *)
         let tt = Term.rename t in
         let parallel_patterns = Listset.remove [] (Listx.power patterns) in
         List.map (make_peak tt lr) parallel_patterns)
       (parallel_overlaps_aux r s))

(* [Rule] -> [Rule] -> [Term,[Pos],Term,Term] *)
let parallel_cpp2 r s =
  Listx.unique
    (List.map
      (fun (t,pats,s,_,u) -> (t,(List.map snd pats),s,u))
      (parallel_overlaps r s))

let parallel_cpp r = parallel_cpp2 r r

let parallel_cp2 r s = List.map (fun (t,_,_,u) -> (t,u)) (parallel_cpp2 r s)

let parallel_cp r = parallel_cp2 r r


(** (weakly) Orthogonality **)

(* mutually orthogonal *)
let is_WMO r s =
  Trs.left_linear (r @ s) &&
  List.for_all (fun (s,t) -> s = t)
    (cp2_all r s)

(* orthogonal *)
let is_OH trs =
  is_WMO trs trs



(**********************************************)
(**************** Comparison ******************)
(**********************************************)
let compare s t =
  if variant s t then
    0
  else
    Stdlib.compare s t
