open Format
open Formatx
open Monad.MonadList

type t = 
  | V of string
  | F of string * t list

type binding = string * t

type position = int list

type subst = binding list

let rec print ppf = function
  | V x -> fprintf ppf "%s" x
  | F (f, ts) -> fprintf ppf "%s(%a)" f (print_list print ",") ts

let sprint t =
  print str_formatter t;
  flush_str_formatter ()

let rec print_subst_aux ppf (x,t) =
  fprintf ppf "%s |-> %a" x print t

let print_subst ppf sigma =
  fprintf ppf "{@.%a@.}@." (print_list print_subst_aux "\n") sigma

let print_position ppf p =
  fprintf ppf "[%a]" (print_list pp_print_int ",") p

let rec functions_list = function
  | V _ -> []
  | F (f, ts) ->
      [f] @ Listx.concat_map functions_list ts

let functions t = Listx.unique (functions_list t)

let rec variables_list = function
  | V x -> [x]
  | F (f, ts) ->
      Listx.concat_map variables_list ts

let variables t = Listx.unique (variables_list t)

let rec count_variable x = function
  | V y when x = y -> 1
  | V _ -> 0
  | F (f, ts) ->
      Listx.sum (List.map (count_variable x) ts)

let term_size t = List.length (functions_list t @ variables_list t)

let is_variable = function
  | V _ -> true
  | F (_, _) -> false

let rec signature = function
  | V _ -> []
  | F (f, ts) -> 
      let sigs = Listx.concat_map signature ts in
      Listx.unique ((f, List.length ts) :: sigs)

let subterms_with_positions t =
  let rec loop t p = match t with
    | V _      -> [t,p]
    | F (_,[]) -> [t,p]
    | F (_,ts) -> (t,p) ::
        Listx.concat_map
          (fun (i,x) -> loop x (p@[i]))
          (Listx.ix ts)
  in loop t []

let rec subterms = function
  | V _ as t -> [t]
  | F (_, ts) as t -> 
      t :: Listx.concat_map subterms (Listx.unique ts)

let proper_subterms t = Listset.diff (subterms t) [t]

let root = function
  | V x -> invalid_arg "Term.root"
  | F (f, _) -> f

let is_subterm s t = List.mem s (subterms t)

let is_proper_subterm s t = List.mem s (proper_subterms t)

let apply subst x = try List.assoc x subst with Not_found -> V x

let rec substitute subst = function
  | V x -> apply subst x
  | F (f, ts) ->
      let ss = List.map (substitute subst) ts in
      F (f, ss)


let count = ref 0

let fresh () = incr count; sprintf "_y%d" !count

let new_var () = V (fresh ())

let new_const () = F (fresh (),[])

let rec ren = function
  | V x -> V (fresh ())
  | F (f, ts) -> F (f, List.map ren ts)

let rename l =
  let xs = variables l in
  let subst =
    List.map
      (fun (i,x) -> x, V (sprintf "$%d" i))
      (Listx.ix xs) in
  substitute subst l

let direct_subterms = function
  | V _ -> []
  | F (_, ss) -> ss

let rec subterm_at p t =
  match p, t with
  | [], _ -> t
  | i :: q, F (_, ts) -> subterm_at q (List.nth ts i)
  | _ -> invalid_arg "Term.subterm_at"

let rec positions = function
  | V x -> [[]]
  | F (_, ts) ->
      let ps =
        let* i,ti = Listx.ix ts in
        let* q = positions ti in
        [i::q] in
      [] :: ps

let is_subterm_at p t s =
  List.mem p (positions s) && t = subterm_at p s

let subterms_at ps t =
  let qs = Listset.inter ps (positions t) in
  List.map (fun q -> subterm_at q t) qs

let variable_positions t =
  List.filter
    (fun p -> is_variable (subterm_at p t))
    (positions t)

let function_positions t =
  Listset.diff (positions t) (variable_positions t)

(* list of s' positions in t *)
let positions_of s t =
  let* (u,p) = subterms_with_positions t in
  if u = s then
    [p]
  else
    []

let variables_at ps s =
  Listx.concat_map variables (subterms_at ps s)

let rec replace s t p =
  match s, p with
  | _, [] -> t
  | F (f, ss), i :: q ->
      let ts =
        List.map
          (fun (j,sj) -> if i = j then replace sj t q else sj)
          (Listx.ix ss) in
      F (f, ts)
  | _ -> invalid_arg "Term.replace"

let rec count_variables t x =
  match t with
  | V y when x = y -> 1
  | V _ -> 0
  | F (_, ts) ->
      let cs = List.map (fun ti -> count_variables ti x) ts in
      List.fold_left (+) 0 cs

let linear t =
  List.for_all (fun x -> count_variables t x = 1) (variables t)

(* for ac *)

let head f = function
  | F(g,gs) when f=g -> true
  | _ -> false

let rec unflatten ac = function
  | V _ as v -> v
  | F(f,ts) ->
     let unflat_args = List.map (unflatten ac) ts in
     if (List.mem f ac) && List.length ts > 2 then
      make_args f unflat_args 
     else F(f,unflat_args)
and make_args f = function
  | h1 :: h2 :: [] -> F(f,[h1;h2])
  | h1 :: h2 :: tt -> 
	F(f,[h1; (make_args f (h2 :: tt))])
  | _ -> invalid_arg "Term.make_args"

let rec flatten ac = function
  | V _ as v -> v
  | F(f,ts) ->
      let ts_flat = List.map (flatten ac) ts in 
      if not (List.mem f ac) then F(f, ts_flat)
      else
	if List.exists (head f) ts 
	then flatten ac (F(f, iter f ts_flat))
	else F(f,ts_flat)
and iter f = function
  | [] -> []
  | (F(g,gs)) :: tt when f=g -> gs @ (iter f tt)
  | h :: tt -> h :: (iter f tt)
