open Format
open Term
open Monad.MonadList

type t = Rule.t list

(** Printitng **)
module E = struct
  type t = Rule.t
  let compare = compare
end
module I = Index.Make (E)

let label rs = List.iter (fun lr -> ignore (I.index lr)) rs

(* OCaml evaluation is started at rhs,
so (index x) > (index (y) for fresh x, y ==> false *)
let sort trs =
    ignore (label trs);
    List.sort (fun a b -> compare (I.index a) (I.index b)) trs

let sprint_rule (l,r) =
  sprintf "  %2d: %s -> %s@." (I.index (l,r)) (sprint l) (sprint r)

let sprint_rules trs = 
  List.fold_left
    (fun s rule -> sprintf "%s%s" s (sprint_rule rule))
    ""
    (sort trs)

let print_rules trs = print_endline (sprint_rules trs)

(*
let sprint_overlap ppf ((a,_),p,(b,_)) =
  sprintf "%s, %s, %s@." (sprint a) (sprint_pos p) (sprint b)

let sprint_overlaps laps = List.fold_left
  (fun s lap -> s ^ sprint_overlap "%s%s" lap)
  ""
  laps

let print_overlaps laps = print_endline (sprint_overlaps laps)

(* a <-^p t -> b *)
let sprint_cp (a,p,t,b) =
  sprintf "%s <- %s, %s -> %s@."
    (sprint a) (sprint_pos p) (sprint t) (sprint b)

let print_cp cp = print_endline (sprint_cp cp)
*)


(********************************************************)

let inverse trs = List.map (fun (l,r) -> (r,l)) trs

(** Symbols **)

let variables trs =
  let vs =
    let* (l,r) = trs in
    Term.variables l @ Term.variables r in
  Listx.unique vs

let functions trs =
  let fs =
    let* (l,r) = trs in
    Term.functions l @ Term.functions r in
  Listx.unique fs

let def (l,_) = root l
let defs trs  = Listx.unique @@ List.map def trs

let cons rule =
  Listset.diff
    (Listx.unique @@ functions [rule])
    [def rule]
let conss trs =
  Listset.diff (functions trs) (defs trs)


(** Subterms **)

let subterms trs =
  let ss =
    let* (l,r) = trs in
    Term.subterms l @ Term.subterms r in
  Listx.unique ss


(** Substitution **)

let subst s trs =
  List.map
    (fun (l,r) -> Term.substitute s l, Term.substitute s r)
    trs

let make_fresh_map vs =
  List.map (fun x -> (x, Term.new_var ())) vs

let rename trs =
  let* (l,r) = trs in
  let s = make_fresh_map (variables [l,r]) in
  [Term.substitute s l, Term.substitute s r]


(** Predicate **)
let trs_cond1 trs =
  List.for_all (fun (l,_) -> not (is_variable l)) trs

let trs_cond2 trs =
  List.for_all (fun (l,r) -> Listset.subset (Term.variables r) (Term.variables l)) trs

let is_trs trs = trs_cond1 trs && trs_cond2 trs

let linear trs =
  List.for_all
    (fun (l,r) -> Term.linear l && Term.linear r)
    trs

let left_linear trs =
  List.for_all
    (fun (l,_) -> Term.linear l)
    trs

let right_linear trs =
  List.for_all
    (fun (_,r) -> Term.linear r)
    trs

let non_erase trs =
  List.for_all
    (fun (l,r) -> Listset.subset (Term.variables l) (Term.variables r))
    trs

let variant r s =
  List.for_all (fun lr -> List.exists (Rule.variant lr) s) r
  &&
  List.for_all (fun lr -> List.exists (Rule.variant lr) r) s

(** Initialize **)
let trs_cond1' trs =
  List.filter (fun lr -> not (trs_cond1 [lr])) trs
let trs_cond2' trs =
  List.filter (fun lr -> not (trs_cond2 [lr])) trs

let init trs =
  let rr = rename trs in
  (* add rule index *)
  ignore (List.map I.index rr);
  (* check validity of TRS *)
  let error1 rs = Format.eprintf
      "ERROR\nleft-hand-side is variable\n%s@."
      (sprint_rules rs)
  and error2 rs = Format.eprintf
      "ERROR\nfree variable exists in right-hand side\n%s@."
      (sprint_rules rs)
  in match trs_cond1' rr, trs_cond2' rr with
  | [],[] -> rr
  | xs,[] -> (error1 xs; exit 1)
  | [],ys -> (error2 ys; exit 1)
  | xs,ys -> (error1 xs; error2 ys; exit 1)
